# lingo - payment and billing system
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import csv

from django import VERSION
from django.core.exceptions import PermissionDenied
from django.db.models import (
    CharField,
    Count,
    DecimalField,
    Exists,
    IntegerField,
    JSONField,
    OuterRef,
    Prefetch,
    Q,
    Subquery,
    Sum,
    Value,
)
from django.db.models.functions import Coalesce
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.template.defaultfilters import yesno
from django.urls import reverse
from django.utils.timezone import localtime, now
from django.utils.translation import gettext_lazy as _
from django.utils.translation import pgettext
from django.views.generic import CreateView, DeleteView, DetailView, FormView, ListView, UpdateView

from lingo.agendas.models import Agenda
from lingo.export_import.views import WithApplicationsMixin
from lingo.invoicing.forms import (
    PaymentDocketForm,
    PaymentDocketPaymentTypeForm,
    PaymentTypeForm,
    RegieCreditFilterSet,
    RegieDocketPaymentFilterSet,
    RegieForm,
    RegieInvoiceCancelForm,
    RegieInvoiceFilterSet,
    RegiePaymentCancelForm,
    RegiePaymentFilterSet,
    RegiePublishingForm,
    RegieRefundFilterSet,
)
from lingo.invoicing.models import (
    PAYMENT_INFO,
    Counter,
    Credit,
    InjectedLine,
    Invoice,
    InvoiceLine,
    InvoiceLinePayment,
    JournalLine,
    Payment,
    PaymentDocket,
    PaymentType,
    Pool,
    Refund,
    Regie,
)
from lingo.invoicing.views.utils import PDFMixin
from lingo.manager.utils import (
    CanBeControlledCheckMixin,
    CanBeInvoicedCheckMixin,
    CanBeManagedCheckMixin,
    CanBeManagedRequiredMixin,
    CanBeViewedCheckMixin,
    CanBeViewedRequiredMixin,
    StaffRequiredMixin,
)
from lingo.snapshot.models import RegieSnapshot
from lingo.snapshot.views import InstanceWithSnapshotHistoryCompareView, InstanceWithSnapshotHistoryView
from lingo.utils.misc import json_dump


class RegiesListView(WithApplicationsMixin, ListView):
    template_name = 'lingo/invoicing/manager_regie_list.html'
    model = Regie

    def dispatch(self, request, *args, **kwargs):
        self.with_applications_dispatch(request)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        queryset = self.with_applications_queryset()
        if not self.request.user.is_staff:
            group_ids = [x.id for x in self.request.user.groups.all()]
            queryset = queryset.filter(
                Q(view_role_id__in=group_ids)
                | Q(edit_role_id__in=group_ids)
                | Q(invoice_role_id__in=group_ids)
                | Q(control_role_id__in=group_ids)
            )
        return queryset

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        return self.with_applications_context_data(context)


regies_list = RegiesListView.as_view()


class RegieAddView(StaffRequiredMixin, CreateView):
    template_name = 'lingo/invoicing/manager_regie_form.html'
    model = Regie
    fields = ['label', 'description', 'edit_role', 'view_role', 'invoice_role', 'control_role']

    def form_valid(self, form):
        response = super().form_valid(form)
        PaymentType.create_defaults(self.object)
        self.object.take_snapshot(request=self.request, comment=pgettext('snapshot', 'created'))
        return response

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-detail', args=[self.object.pk])


regie_add = RegieAddView.as_view()


class RegieDetailView(CanBeViewedRequiredMixin, DetailView):
    template_name = 'lingo/invoicing/manager_regie_detail.html'
    model = Regie

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.object
        kwargs['campaigns'] = self.object.campaign_set.all().order_by('-date_start')
        kwargs['user_can_invoice'] = self.object.can_be_invoiced(self.request.user)
        return super().get_context_data(**kwargs)


regie_detail = RegieDetailView.as_view()


class RegieParametersView(CanBeViewedRequiredMixin, DetailView):
    template_name = 'lingo/invoicing/manager_regie_parameters.html'
    model = Regie

    def get_queryset(self):
        paymenttypes_qs = PaymentType.objects.annotate(
            used=Exists(Payment.objects.filter(payment_type=OuterRef('id')))
        )
        return (
            super()
            .get_queryset()
            .select_related('edit_role', 'view_role', 'invoice_role', 'control_role')
            .prefetch_related(Prefetch('paymenttype_set', queryset=paymenttypes_qs))
        )

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.object
        kwargs['agendas'] = Agenda.objects.filter(regie=self.object)
        has_related_objects = False
        if self.object.campaign_set.exists():
            has_related_objects = True
        elif self.object.injectedline_set.exists():
            has_related_objects = True
        kwargs['has_related_objects'] = has_related_objects
        kwargs['user_can_manage'] = self.object.can_be_managed(self.request.user)
        return super().get_context_data(**kwargs)


regie_parameters = RegieParametersView.as_view()


class RegieEditView(CanBeManagedRequiredMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_regie_form.html'
    model = Regie
    form_class = RegieForm

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-parameters', args=[self.object.pk])

    def form_valid(self, *args, **kwargs):
        response = super().form_valid(*args, **kwargs)
        self.object.take_snapshot(request=self.request)
        return response


regie_edit = RegieEditView.as_view()


class RegiePermissionsEditView(CanBeManagedRequiredMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_regie_permissions_form.html'
    model = Regie
    fields = ['edit_role', 'view_role', 'invoice_role', 'control_role']

    def get_success_url(self):
        return '%s#open:permissions' % reverse(
            'lingo-manager-invoicing-regie-parameters', args=[self.object.pk]
        )

    def form_valid(self, *args, **kwargs):
        response = super().form_valid(*args, **kwargs)
        self.object.take_snapshot(request=self.request, comment=_('changed permissions'))
        return response


regie_permissions_edit = RegiePermissionsEditView.as_view()


class RegieCountersEditView(CanBeManagedRequiredMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_regie_counters_form.html'
    model = Regie
    fields = [
        'counter_name',
        'invoice_number_format',
        'payment_number_format',
        'docket_number_format',
        'credit_number_format',
        'refund_number_format',
    ]

    def get_success_url(self):
        return '%s#open:counters' % reverse('lingo-manager-invoicing-regie-parameters', args=[self.object.pk])

    def form_valid(self, *args, **kwargs):
        response = super().form_valid(*args, **kwargs)
        self.object.take_snapshot(request=self.request, comment=_('changed counters'))
        return response


regie_counters_edit = RegieCountersEditView.as_view()


class RegiePublishingEditView(CanBeManagedRequiredMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_regie_publishing_form.html'
    model = Regie
    form_class = RegiePublishingForm

    def get_success_url(self):
        return '%s#open:publishing' % reverse(
            'lingo-manager-invoicing-regie-parameters', args=[self.object.pk]
        )

    def form_valid(self, *args, **kwargs):
        response = super().form_valid(*args, **kwargs)
        self.object.take_snapshot(request=self.request, comment=_('changed publishing settings'))
        return response


regie_publishing_edit = RegiePublishingEditView.as_view()


class RegieDeleteView(CanBeManagedRequiredMixin, DeleteView):
    template_name = 'lingo/manager_confirm_delete.html'
    model = Regie

    def get_queryset(self):
        return super().get_queryset().filter(campaign__isnull=True, injectedline__isnull=True)

    def delete(self, request, *args, **kwargs):
        if VERSION < (4, 0):
            self.post_delete_actions()
            return super().delete(request, *args, **kwargs)

    def form_valid(self, form):
        if VERSION >= (4, 0):
            self.post_delete_actions()
            return super().form_valid(form)

    def post_delete_actions(self):
        self.object = self.get_object()
        self.object.take_snapshot(request=self.request, deletion=True)
        Counter.objects.filter(regie=self.object).delete()
        PaymentType.objects.filter(regie=self.object).delete()

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-list')


regie_delete = RegieDeleteView.as_view()


class RegieExport(CanBeManagedRequiredMixin, DetailView):
    model = Regie

    def get(self, request, *args, **kwargs):
        response = HttpResponse(content_type='application/json')
        attachment = 'attachment; filename="export_regie_{}_{}.json"'.format(
            self.get_object().slug, now().strftime('%Y%m%d')
        )
        response['Content-Disposition'] = attachment
        json_dump({'regies': [self.get_object().export_json()]}, response, indent=2)
        return response


regie_export = RegieExport.as_view()


class PaymentTypeAddView(CanBeManagedCheckMixin, CreateView):
    template_name = 'lingo/invoicing/manager_payment_type_form.html'
    model = PaymentType
    fields = ['label']

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs.pop('regie_pk'))
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        if not kwargs.get('instance'):
            kwargs['instance'] = self.model()
        kwargs['instance'].regie = self.regie
        return kwargs

    def get_success_url(self):
        return '%s#open:payment-types' % reverse(
            'lingo-manager-invoicing-regie-parameters', args=[self.regie.pk]
        )

    def form_valid(self, *args, **kwargs):
        response = super().form_valid(*args, **kwargs)
        self.regie.take_snapshot(request=self.request, comment=_('added payment type'))
        return response


payment_type_add = PaymentTypeAddView.as_view()


class PaymentTypeEditView(CanBeManagedCheckMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_payment_type_form.html'
    model = PaymentType
    form_class = PaymentTypeForm

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs.pop('regie_pk'))
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_queryset(self):
        return PaymentType.objects.filter(regie=self.regie)

    def get_success_url(self):
        return '%s#open:payment-types' % reverse(
            'lingo-manager-invoicing-regie-parameters', args=[self.regie.pk]
        )

    def form_valid(self, *args, **kwargs):
        response = super().form_valid(*args, **kwargs)
        self.regie.take_snapshot(request=self.request, comment=_('changed payment type'))
        return response


payment_type_edit = PaymentTypeEditView.as_view()


class PaymentTypeDeleteView(CanBeManagedCheckMixin, DeleteView):
    template_name = 'lingo/manager_confirm_delete.html'
    model = PaymentType

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs.pop('regie_pk'))
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return PaymentType.objects.filter(regie=self.regie.pk, payment__isnull=True)

    def get_success_url(self):
        return '%s#open:payment-types' % reverse(
            'lingo-manager-invoicing-regie-parameters', args=[self.regie.pk]
        )

    def post(self, *args, **kwargs):
        response = super().post(*args, **kwargs)
        self.regie.take_snapshot(request=self.request, comment=_('removed payment type'))
        return response


payment_type_delete = PaymentTypeDeleteView.as_view()


class NonInvoicedLineListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_non_invoiced_line_list.html'
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        fields = [
            'pk',
            'event_date',
            'slug',
            'label',
            'amount',
            'user_external_id',
            'payer_external_id',
            'payer_first_name',
            'payer_last_name',
            'user_first_name',
            'user_last_name',
            'event',
            'pricing_data',
            'status',
            'pool_id',
        ]
        qs1 = JournalLine.objects.filter(
            status='error', error_status='', pool__campaign__regie=self.regie
        ).values(*fields)
        qs2 = (
            InjectedLine.objects.filter(journalline__isnull=True, regie=self.regie)
            .annotate(
                user_first_name=Value('', output_field=CharField()),
                user_last_name=Value('', output_field=CharField()),
                event=Value({}, output_field=JSONField()),
                pricing_data=Value({}, output_field=JSONField()),
                status=Value('injected', output_field=CharField()),
                pool_id=Value(0, output_field=IntegerField()),
            )
            .values(*fields)
        )
        qs = qs1.union(qs2).order_by('event_date', 'user_external_id', 'label', 'pk')
        return qs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        context = super().get_context_data(**kwargs)
        pools = Pool.objects.filter(draft=False).in_bulk()
        for line in context['object_list']:
            if line['status'] == 'error':
                line['user_name'] = JournalLine(
                    user_first_name=line['user_first_name'], user_last_name=line['user_last_name']
                ).user_name
                line['payer_name'] = JournalLine(
                    payer_first_name=line['payer_first_name'], payer_last_name=line['payer_last_name']
                ).payer_name
                line['error_display'] = JournalLine(
                    status=line['status'], pricing_data=line['pricing_data']
                ).get_error_display()
                line['campaign_id'] = pools[line['pool_id']].campaign_id
                line['chrono_event_url'] = JournalLine(event=line['event']).get_chrono_event_url()
        return context


non_invoiced_line_list = NonInvoicedLineListView.as_view()


class RegieInvoiceListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_invoice_list.html'
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        if 'csv' in request.GET and not self.regie.can_be_invoiced(self.request.user):
            raise PermissionDenied()
        self.full = bool('full' in request.GET)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        invoice_queryset = (
            Invoice.objects.filter(
                Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
                regie=self.regie,
            )
            .prefetch_related('pool')
            .order_by('-created_at')
        )
        if self.full:
            payment_queryset = InvoiceLinePayment.objects.select_related('payment__payment_type')
            invoice_queryset = invoice_queryset.prefetch_related(
                Prefetch('lines', queryset=InvoiceLine.objects.all().order_by('pk')),
                Prefetch('lines__invoicelinepayment_set', queryset=payment_queryset),
            )
        self.filterset = RegieInvoiceFilterSet(data=self.request.GET or None, queryset=invoice_queryset)
        return self.filterset.qs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['filterset'] = self.filterset
        kwargs['user_can_invoice'] = self.regie.can_be_invoiced(self.request.user)
        return super().get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        if 'csv' in request.GET and self.filterset.form.is_valid():
            return self.csv(request, context)
        return self.render_to_response(context)

    def csv(self, request, context):
        response = HttpResponse(content_type='text/csv')
        if self.full:
            response['Content-Disposition'] = 'attachment; filename="invoices-full.csv"'
        else:
            response['Content-Disposition'] = 'attachment; filename="invoices.csv"'
        writer = csv.writer(response)
        headers = [
            _('Number'),
            _('Payer ID'),
            _('Payer first name'),
            _('Payer last name'),
            _('Publication date'),
            _('Payment deadline'),
            _('Due date'),
            _('Demat'),
            _('Direct debit'),
        ]
        if self.full:
            headers += [
                _('Description'),
                _('Accounting code'),
                _('Unit amount'),
                _('Quantity'),
            ]
        headers += [
            _('Total due'),
        ]
        if self.full:
            headers += [
                _('Payment type'),
            ]
        headers += [
            _('Paid amount'),
            _('Status'),
            _('Cancelled on'),
            _('Cancellation reason'),
        ]
        # headers
        writer.writerow(headers)
        for invoice in self.object_list:
            paid_status = _('Not paid')
            if invoice.cancelled_at is not None:
                paid_status = pgettext('invoice', 'Cancelled')
            elif invoice.remaining_amount > 0 and invoice.paid_amount > 0:
                paid_status = _('Partially paid')
            elif invoice.remaining_amount == 0:
                paid_status = _('Paid')
            if not self.full:
                writer.writerow(
                    [
                        invoice.formatted_number,
                        invoice.payer_external_id,
                        invoice.payer_first_name,
                        invoice.payer_last_name,
                        invoice.date_publication.isoformat(),
                        invoice.date_payment_deadline.isoformat(),
                        invoice.date_due.isoformat(),
                        yesno(invoice.payer_demat),
                        yesno(invoice.payer_direct_debit),
                        invoice.total_amount,
                        invoice.paid_amount,
                        paid_status,
                        invoice.cancelled_at.isoformat() if invoice.cancelled_at else '',
                        invoice.cancellation_reason or '',
                    ]
                )
                continue
            for line in invoice.lines.all():
                if line.total_amount == 0:
                    continue
                paid_status = _('Not paid')
                if invoice.cancelled_at is not None:
                    paid_status = pgettext('invoice', 'Cancelled')
                elif line.remaining_amount > 0 and line.paid_amount > 0:
                    paid_status = _('Partially paid')
                elif line.remaining_amount == 0:
                    paid_status = _('Paid')
                payment_types = {p.payment.payment_type.label for p in line.invoicelinepayment_set.all()}
                writer.writerow(
                    [
                        invoice.formatted_number,
                        invoice.payer_external_id,
                        invoice.payer_first_name,
                        invoice.payer_last_name,
                        invoice.date_publication.isoformat(),
                        invoice.date_payment_deadline.isoformat(),
                        invoice.date_due.isoformat(),
                        yesno(invoice.payer_demat),
                        yesno(invoice.payer_direct_debit),
                        line.label,
                        line.accounting_code,
                        line.unit_amount,
                        line.quantity,
                        line.total_amount,
                        ', '.join(sorted(payment_types)),
                        line.paid_amount,
                        paid_status,
                        invoice.cancelled_at.isoformat() if invoice.cancelled_at else '',
                        invoice.cancellation_reason or '',
                    ]
                )
        return response


regie_invoice_list = RegieInvoiceListView.as_view()


class RegieInvoicePDFView(CanBeViewedCheckMixin, PDFMixin, DetailView):
    pk_url_kwarg = 'invoice_pk'
    model = Invoice

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
                regie=self.regie,
            )
        )


regie_invoice_pdf = RegieInvoicePDFView.as_view()


class RegieInvoicePaymentsPDFView(CanBeViewedCheckMixin, PDFMixin, DetailView):
    pk_url_kwarg = 'invoice_pk'
    model = Invoice

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
                regie=self.regie,
                remaining_amount=0,
            )
        )

    def html(self):
        return self.object.payments_html()

    def get_filename(self):
        return 'A-%s' % self.object.formatted_number


regie_invoice_payments_pdf = RegieInvoicePaymentsPDFView.as_view()


class RegieInvoiceLineListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_invoice_lines.html'

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        self.invoice = get_object_or_404(
            Invoice,
            Q(pool__isnull=True) | Q(pool__campaign__finalized=True),
            pk=kwargs['invoice_pk'],
            regie=self.regie,
        )
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.invoice.get_grouped_and_ordered_lines()

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['invoice'] = self.invoice
        kwargs['user_can_invoice'] = self.regie.can_be_invoiced(self.request.user)
        return super().get_context_data(**kwargs)


regie_invoice_line_list = RegieInvoiceLineListView.as_view()


class RegieInvoiceCancelView(CanBeInvoicedCheckMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_invoice_cancel_form.html'
    pk_url_kwarg = 'invoice_pk'
    model = Invoice
    form_class = RegieInvoiceCancelForm

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                regie=self.regie,
                cancelled_at__isnull=True,
            )
            .exclude(pk__in=InvoiceLine.objects.filter(invoicelinepayment__isnull=False).values('invoice'))
        )

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return '%s?number=%s' % (
            reverse('lingo-manager-invoicing-regie-invoice-list', args=[self.regie.pk]),
            self.object.formatted_number,
        )


regie_invoice_cancel = RegieInvoiceCancelView.as_view()


class RegiePaymentListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_payment_list.html'
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        if 'csv' in request.GET and not self.regie.can_be_controlled(self.request.user):
            raise PermissionDenied()
        self.full = bool('full' in request.GET)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        invoice_line_payment_queryset = InvoiceLinePayment.objects.select_related('line__invoice').order_by(
            'created_at'
        )
        self.filterset = RegiePaymentFilterSet(
            data=self.request.GET or None,
            queryset=Payment.objects.filter(regie=self.regie)
            .prefetch_related(
                'payment_type',
                Prefetch(
                    'invoicelinepayment_set',
                    queryset=invoice_line_payment_queryset,
                    to_attr='prefetched_invoicelinepayments',
                ),
                'docket',
            )
            .order_by('-created_at'),
            regie=self.regie,
        )
        return self.filterset.qs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['filterset'] = self.filterset
        kwargs['user_can_control'] = self.regie.can_be_controlled(self.request.user)
        return super().get_context_data(**kwargs)

    def get(self, request, *args, **kwargs):
        self.object_list = self.get_queryset()
        context = self.get_context_data()
        if 'csv' in request.GET and self.filterset.form.is_valid():
            return self.csv(request, context)
        return self.render_to_response(context)

    def csv(self, request, context):
        response = HttpResponse(content_type='text/csv')
        if self.full:
            response['Content-Disposition'] = 'attachment; filename="payments_full.csv"'
        else:
            response['Content-Disposition'] = 'attachment; filename="payments.csv"'
        writer = csv.writer(response)
        # headers
        headers = [
            _('Number'),
            _('Invoice number'),
            _('Date'),
            _('Payer ID'),
            _('Payer first name'),
            _('Payer last name'),
        ]
        if self.full:
            headers += [
                _('Description (line)'),
                _('Accounting code (line)'),
                _('Amount (line)'),
                _('Quantity (line)'),
                _('Subtotal (line)'),
                _('Amount assigned (line)'),
            ]
        headers += (
            [
                _('Payment type'),
                _('Amount assigned (invoice)'),
                _('Total amount (payment)'),
            ]
            + [v for k, v in PAYMENT_INFO]
            + [
                _('Debt reference'),
                _('Cancelled on'),
                _('Cancellation reason'),
            ]
        )
        writer.writerow(headers)
        for payment in self.object_list:
            invoice_payments = payment.get_invoice_payments()
            for invoice_payment in invoice_payments:
                if not self.full:
                    writer.writerow(
                        [
                            payment.formatted_number,
                            invoice_payment.invoice.formatted_number,
                            localtime(payment.created_at).date().isoformat(),
                            payment.payer_external_id,
                            payment.payer_first_name,
                            payment.payer_last_name,
                            payment.payment_type.label,
                            invoice_payment.amount,
                            payment.amount,
                        ]
                        + [payment.payment_info.get(k) or '' for k, v in PAYMENT_INFO]
                        + [
                            payment.bank_data.get('refdet') or '',
                            payment.cancelled_at.isoformat() if payment.cancelled_at else '',
                            payment.cancellation_reason or '',
                        ]
                    )
                    continue
                for invoice_payment_line in invoice_payment.lines:
                    writer.writerow(
                        [
                            payment.formatted_number,
                            invoice_payment.invoice.formatted_number,
                            localtime(payment.created_at).date().isoformat(),
                            payment.payer_external_id,
                            payment.payer_first_name,
                            payment.payer_last_name,
                            invoice_payment_line.line.label,
                            invoice_payment_line.line.accounting_code,
                            invoice_payment_line.line.unit_amount,
                            invoice_payment_line.line.quantity,
                            invoice_payment_line.line.total_amount,
                            invoice_payment_line.amount,
                            payment.payment_type.label,
                            invoice_payment.amount,
                            payment.amount,
                        ]
                        + [payment.payment_info.get(k) or '' for k, v in PAYMENT_INFO]
                        + [
                            payment.bank_data.get('refdet') or '',
                            payment.cancelled_at.isoformat() if payment.cancelled_at else '',
                            payment.cancellation_reason or '',
                        ]
                    )
            if not invoice_payments:
                row = [
                    payment.formatted_number,
                    '',
                    localtime(payment.created_at).date().isoformat(),
                    payment.payer_external_id,
                    payment.payer_first_name,
                    payment.payer_last_name,
                ]
                if self.full:
                    row += ['', '', '', '', '', '']
                row += [
                    payment.payment_type.label,
                    '',
                    payment.amount,
                ]
                row += [payment.payment_info.get(k) or '' for k, v in PAYMENT_INFO]
                row += [
                    payment.bank_data.get('refdet') or '',
                    payment.cancelled_at.isoformat() if payment.cancelled_at else '',
                    payment.cancellation_reason or '',
                ]
                writer.writerow(row)
        return response


regie_payment_list = RegiePaymentListView.as_view()


class RegiePaymentPDFView(CanBeViewedCheckMixin, PDFMixin, DetailView):
    pk_url_kwarg = 'payment_pk'
    model = Payment

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return super().get_queryset().filter(regie=self.regie)


regie_payment_pdf = RegiePaymentPDFView.as_view()


class RegiePaymentCancelView(CanBeControlledCheckMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_payment_cancel_form.html'
    pk_url_kwarg = 'payment_pk'
    model = Payment
    form_class = RegiePaymentCancelForm

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return super().get_queryset().filter(regie=self.regie, cancelled_at__isnull=True)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['request'] = self.request
        return kwargs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return '%s?number=%s' % (
            reverse('lingo-manager-invoicing-regie-payment-list', args=[self.regie.pk]),
            self.object.formatted_number,
        )


regie_payment_cancel = RegiePaymentCancelView.as_view()


class RegieDocketPaymentListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_docket_payment_list.html'
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        self.filterset = RegieDocketPaymentFilterSet(
            data=self.request.GET or None,
            queryset=Payment.objects.filter(regie=self.regie, docket__isnull=True, cancelled_at__isnull=True)
            .select_related('payment_type')
            .order_by('-created_at'),
            regie=self.regie,
        )
        if not self.filterset.form.is_valid():
            return Payment.objects.none()
        return self.filterset.qs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['filterset'] = self.filterset
        kwargs['has_draft'] = self.regie.paymentdocket_set.filter(draft=True).exists()
        kwargs['user_can_control'] = self.regie.can_be_controlled(self.request.user)
        return super().get_context_data(**kwargs)


regie_docket_payment_list = RegieDocketPaymentListView.as_view()


class RegieDocketListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_docket_list.html'
    model = PaymentDocket
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        payments = Payment.objects.filter(docket=OuterRef('pk')).order_by().values('docket')
        active_count = (
            payments.filter(cancelled_at__isnull=True).annotate(count=Count('docket')).values('count')
        )
        active_amount = (
            payments.filter(cancelled_at__isnull=True).annotate(total=Sum('amount')).values('total')
        )
        cancelled_count = (
            payments.filter(cancelled_at__isnull=False).annotate(count=Count('docket')).values('count')
        )
        cancelled_amount = (
            payments.filter(cancelled_at__isnull=False).annotate(total=Sum('amount')).values('total')
        )
        return (
            self.regie.paymentdocket_set.all()
            .prefetch_related('payment_types')
            .annotate(
                active_count=Coalesce(Subquery(active_count, output_field=IntegerField()), Value(0)),
                cancelled_count=Coalesce(Subquery(cancelled_count, output_field=IntegerField()), Value(0)),
                active_amount=Coalesce(
                    Subquery(active_amount, output_field=DecimalField(max_digits=9, decimal_places=2)),
                    Value(0),
                    output_field=DecimalField(max_digits=9, decimal_places=2),
                ),
                cancelled_amount=Coalesce(
                    Subquery(cancelled_amount, output_field=DecimalField(max_digits=9, decimal_places=2)),
                    Value(0),
                    output_field=DecimalField(max_digits=9, decimal_places=2),
                ),
            )
            .order_by('-created_at')
        )

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)


regie_docket_list = RegieDocketListView.as_view()


class RegieDocketAddView(CanBeControlledCheckMixin, CreateView):
    template_name = 'lingo/invoicing/manager_docket_form.html'
    model = PaymentDocket
    form_class = PaymentDocketForm

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        if self.regie.paymentdocket_set.filter(draft=True).exists():
            raise Http404
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        if not kwargs.get('instance'):
            kwargs['instance'] = self.model()
        kwargs['instance'].regie = self.regie
        kwargs['instance'].draft = True
        kwargs['regie'] = self.regie
        return kwargs

    def get(self, request, *args, **kwargs):
        form = self.get_form_class()(
            data={
                'payment_types': request.GET.getlist('payment_type'),
                'date_end': request.GET.get('date_end'),
            }
            if request.GET
            else None,
            **self.get_form_kwargs(),
        )
        if form.is_valid():
            return self.form_valid(form)
        return super().get(request, *args, **kwargs)

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-docket-detail', args=[self.regie.pk, self.object.pk])


regie_docket_add = RegieDocketAddView.as_view()


class RegieDocketDetailView(CanBeViewedCheckMixin, DetailView):
    template_name = 'lingo/invoicing/manager_docket_detail.html'
    model = PaymentDocket

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.regie.paymentdocket_set.all()

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['active'] = self.object.get_active_payments()
        kwargs['cancelled'] = self.object.get_cancelled_payments()
        kwargs['user_can_control'] = self.regie.can_be_controlled(self.request.user)
        return super().get_context_data(**kwargs)


regie_docket_detail = RegieDocketDetailView.as_view()


class RegieDocketCSVView(CanBeControlledCheckMixin, DetailView):
    model = PaymentDocket

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.regie.paymentdocket_set.all()

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="docket-%s.csv"' % self.object
        writer = csv.writer(response)
        writer.writerow(
            [
                _('Initial docket amount'),
                _('Cancelled docket amount'),
                _('Final docket amount'),
            ]
        )
        writer.writerow(
            [
                self.object.get_payments_amount(),
                self.object.get_cancelled_payments_amount(),
                self.object.get_active_payments_amount(),
            ]
        )
        for value in self.object.get_active_payments():
            if not value['list']:
                continue
            writer.writerow(
                [
                    _('Number of payments'),
                    _('Total amount'),
                    _('Payment type'),
                    _('Additional information'),
                ]
            )
            writer.writerow(
                [
                    len(value['list']),
                    value['amount']['amount'],
                    value['payment_type'],
                    self.object.payment_types_info.get(value['payment_type'].slug) or '',
                ]
            )
            headers = [
                _('Number'),
                _('Date'),
                _('Payer ID'),
                _('Payer first name'),
                _('Payer last name'),
                _('Payment type'),
                _('Total amount'),
            ] + [v for k, v in PAYMENT_INFO]
            writer.writerow(headers)
            for payment in value['list']:
                writer.writerow(
                    [
                        payment.formatted_number,
                        localtime(payment.created_at).date().isoformat(),
                        payment.payer_external_id,
                        payment.payer_first_name,
                        payment.payer_last_name,
                        payment.payment_type.label,
                        payment.amount,
                    ]
                    + [payment.payment_info.get(k) or '' for k, v in PAYMENT_INFO]
                )
        cancelled = self.object.get_cancelled_payments()
        if cancelled['list']:
            writer.writerow(
                [
                    _('Number of payments'),
                    _('Total amount'),
                ]
            )
            writer.writerow(
                [
                    len(cancelled['list']),
                    value['amount']['amount'],
                ]
            )
            headers = (
                [
                    _('Number'),
                    _('Date'),
                    _('Payer ID'),
                    _('Payer first name'),
                    _('Payer last name'),
                    _('Payment type'),
                    _('Total amount'),
                ]
                + [v for k, v in PAYMENT_INFO]
                + [
                    _('Cancelled on'),
                    _('Cancellation reason'),
                ]
            )
            writer.writerow(headers)
            for payment in cancelled['list']:
                writer.writerow(
                    [
                        payment.formatted_number,
                        localtime(payment.created_at).date().isoformat(),
                        payment.payer_external_id,
                        payment.payer_first_name,
                        payment.payer_last_name,
                        payment.payment_type.label,
                        payment.amount,
                    ]
                    + [payment.payment_info.get(k) or '' for k, v in PAYMENT_INFO]
                    + [
                        payment.cancelled_at.isoformat(),
                        payment.cancellation_reason or '',
                    ]
                )
        return response


regie_docket_csv = RegieDocketCSVView.as_view()


class RegieDocketPDFView(CanBeControlledCheckMixin, PDFMixin, DetailView):
    model = PaymentDocket

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.regie.paymentdocket_set.all()

    def get_filename(self):
        return self.object


regie_docket_pdf = RegieDocketPDFView.as_view()


class RegieDocketEditView(CanBeControlledCheckMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_docket_form.html'
    model = PaymentDocket
    form_class = PaymentDocketForm

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.regie.paymentdocket_set.filter(draft=True)

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-docket-detail', args=[self.regie.pk, self.object.pk])


regie_docket_edit = RegieDocketEditView.as_view()


class RegieDocketPaymentTypeEditView(CanBeControlledCheckMixin, UpdateView):
    template_name = 'lingo/invoicing/manager_docket_payment_type_form.html'
    model = PaymentDocket
    form_class = PaymentDocketPaymentTypeForm

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.regie.paymentdocket_set.filter(draft=True)

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        return super().get_context_data(**kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['payment_type'] = get_object_or_404(
            self.object.payment_types, pk=self.kwargs['payment_type_pk']
        )
        return kwargs

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-docket-detail', args=[self.regie.pk, self.object.pk])


regie_docket_payment_type_edit = RegieDocketPaymentTypeEditView.as_view()


class RegieDocketValidateView(CanBeControlledCheckMixin, FormView):
    template_name = 'lingo/invoicing/manager_docket_validate.html'

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        self.object = get_object_or_404(
            PaymentDocket,
            regie=self.regie,
            pk=kwargs['pk'],
            draft=True,
        )
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        kwargs['form'] = None
        kwargs['regie'] = self.regie
        kwargs['object'] = self.object
        return super().get_context_data(**kwargs)

    def post(self, request, *args, **kwargs):
        self.object.draft = False
        self.object.set_number()
        self.object.save()
        return redirect(
            reverse('lingo-manager-invoicing-regie-docket-detail', args=[self.regie.pk, self.object.pk])
        )


regie_docket_validate = RegieDocketValidateView.as_view()


class RegieDocketDeleteView(CanBeControlledCheckMixin, DeleteView):
    template_name = 'lingo/manager_confirm_delete.html'
    model = PaymentDocket

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.regie.paymentdocket_set.filter(draft=True)

    def delete(self, request, *args, **kwargs):
        if VERSION < (4, 0):
            self.post_delete_actions()
            return super().delete(request, *args, **kwargs)

    def form_valid(self, form):
        if VERSION >= (4, 0):
            self.post_delete_actions()
            return super().form_valid(form)

    def post_delete_actions(self):
        self.object = self.get_object()
        self.object.payment_set.update(docket=None)

    def get_success_url(self):
        return reverse('lingo-manager-invoicing-regie-docket-list', args=[self.regie.pk])


regie_docket_delete = RegieDocketDeleteView.as_view()


class RegieCreditListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_credit_list.html'
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        credit_queryset = Credit.objects.filter(regie=self.regie).order_by('-created_at')
        self.filterset = RegieCreditFilterSet(data=self.request.GET or None, queryset=credit_queryset)
        return self.filterset.qs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['filterset'] = self.filterset
        return super().get_context_data(**kwargs)


regie_credit_list = RegieCreditListView.as_view()


class RegieCreditLineListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_credit_lines.html'

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        self.credit = get_object_or_404(
            Credit,
            pk=kwargs['credit_pk'],
            regie=self.regie,
        )
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return self.credit.get_grouped_and_ordered_lines()

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['credit'] = self.credit
        return super().get_context_data(**kwargs)


regie_credit_line_list = RegieCreditLineListView.as_view()


class RegieCreditPDFView(CanBeViewedCheckMixin, PDFMixin, DetailView):
    pk_url_kwarg = 'credit_pk'
    model = Credit

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(
                regie=self.regie,
            )
        )


regie_credit_pdf = RegieCreditPDFView.as_view()


class RegieRefundListView(CanBeViewedCheckMixin, ListView):
    template_name = 'lingo/invoicing/manager_refund_list.html'
    paginate_by = 100

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['regie_pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        refund_queryset = Refund.objects.filter(regie=self.regie).order_by('-created_at')
        self.filterset = RegieRefundFilterSet(data=self.request.GET or None, queryset=refund_queryset)
        return self.filterset.qs

    def get_context_data(self, **kwargs):
        kwargs['regie'] = self.regie
        kwargs['filterset'] = self.filterset
        return super().get_context_data(**kwargs)


regie_refund_list = RegieRefundListView.as_view()


class RegieInspectView(CanBeManagedRequiredMixin, DetailView):
    template_name = 'lingo/invoicing/manager_regie_inspect.html'
    model = Regie

    def get_queryset(self):
        return super().get_queryset().select_related('edit_role', 'view_role', 'invoice_role', 'control_role')


regie_inspect = RegieInspectView.as_view()


class RegieHistoryView(CanBeManagedCheckMixin, InstanceWithSnapshotHistoryView):
    template_name = 'lingo/invoicing/manager_regie_history.html'
    model = RegieSnapshot
    instance_context_key = 'regie'

    def dispatch(self, request, *args, **kwargs):
        self.regie = get_object_or_404(Regie, pk=kwargs['pk'])
        self.check_object(self.regie)
        return super().dispatch(request, *args, **kwargs)


regie_history = RegieHistoryView.as_view()


class RegieHistoryCompareView(CanBeManagedRequiredMixin, InstanceWithSnapshotHistoryCompareView):
    template_name = 'lingo/invoicing/manager_regie_history_compare.html'
    inspect_template_name = 'lingo/invoicing/manager_regie_inspect_fragment.html'
    model = Regie
    instance_context_key = 'regie'
    history_view = 'lingo-manager-regie-history'


regie_history_compare = RegieHistoryCompareView.as_view()
